/*
 * The contents of this file are subject to the terms 
 * of the Common Development and Distribution License 
 * (the License).  You may not use this file except in
 * compliance with the License.
 * 
 * You can obtain a copy of the license at 
 * https://glassfish.dev.java.net/public/CDDLv1.0.html or
 * glassfish/bootstrap/legal/CDDLv1.0.txt.
 * See the License for the specific language governing 
 * permissions and limitations under the License.
 * 
 * When distributing Covered Code, include this CDDL 
 * Header Notice in each file and include the License file 
 * at glassfish/bootstrap/legal/CDDLv1.0.txt.  
 * If applicable, add the following below the CDDL Header, 
 * with the fields enclosed by brackets [] replaced by
 * you own identifying information: 
 * "Portions Copyrighted [year] [name of copyright owner]"
 * 
 * Copyright 2006 Sun Microsystems, Inc. All rights reserved.
 */

package org.glassfish.scripting.rails;

import com.sun.enterprise.module.bootstrap.StartupContext;
import com.sun.grizzly.tcp.Request;
import com.sun.grizzly.tcp.Response;
import org.glassfish.api.admin.ServerEnvironment;
import org.glassfish.api.container.Adapter;
import org.glassfish.api.deployment.ApplicationContainer;

import java.io.File;
import java.io.FilenameFilter;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;


/**
 * This is the main Rails application class. It is created from {@link RailsDeployer} during 
 * {@link org.glassfish.api.deployment.Deployer#load(org.glassfish.api.container.Container, org.glassfish.api.deployment.DeploymentContext)}.
 *
 * RailsApplication composes with {@link RailsAdapter}. RailsAdapter create and maintains the JRuby runtime and manages Ruby runtime pool.
 * This basically means that each Rails application will have it's own JRuby runtime object pool. This was the only way to manage the
 * JRuby runtime pool, ideally it could be per container but due to the fact that loading application specific Rails environment thru JRuby
 * is quite expensive.
 *
 * @author vivek Pandey
 */
public class RailsApplication implements ApplicationContainer, Adapter {

    private final Logger logger;
    private Object jRubyRuntime;
    private com.sun.grizzly.tcp.Adapter adapter;

    private final String contextRoot;
    private final JRubyEnvironment jrubyEnv;
    private final ServerEnvironment env;



    RailsApplication(ServerEnvironment env, String railsRoot, Logger logger, String ctxRoot){
        this.logger = logger;
        this.jrubyEnv = new JRubyEnvironment(env.getStartupContext());
        this.env = env;
        this.contextRoot = ctxRoot;
        try {
            /**
             * JRuby is not a OSGi bundle so it needs to be passed custom class loader that makes all the jruby jars availble and also
             * can load any of the java classes available inside the context classloader.
             *
             * So we have a jruby runtime specific wrapper class that weload using custom classloader and this classloader will eventually
             * be used to load everyting else by jruby.
             */
            URL[] urls = findJRubyJars();
            ClassLoader cl = new JRubyRailsClassLoader(urls, RailsApplication.class.getClassLoader());

            Class<?> c = cl.loadClass("com.sun.grizzly.jruby.RubyRuntime");

            this.jRubyRuntime =  c.getConstructor(String.class, String.class, String.class, int.class, boolean.class).newInstance(contextRoot, railsRoot, jrubyEnv.jrubyLib, jrubyEnv.numberOfRuntime, true);
            this.adapter = (com.sun.grizzly.tcp.Adapter) c.getMethod("getAdapter").invoke(jRubyRuntime);

        } catch (Exception e) {
            logger.log(Level.SEVERE, "Error creating RailsAdapter: "+e.getMessage(), e);
            return;
        }
    }

    /**
     * Starts an application container.
     * ContractProvider starting should not throw an exception but rather should
     * use their prefered Logger instance to log any issue they encounter while
     * starting. Returning false from a start mean that the container failed
     * to start and the application that wanted to use the container should
     * no attempt to load().
     *
     * @param cl the application classloader
     * @return true if the container startup was successful.
     */
    public boolean start(ClassLoader cl) {
        return true;
    }

    /**
     * Stop the application container
     * ContractProvider should manage the number of outstanding ContractProvider instances are
     * started and free up resources when that number dip to zero.
     *
     * @return true if stopping was successful.
     */
    public boolean stop() {
        try {
            jRubyRuntime.getClass().getMethod("stopRubyRuntimePool").invoke(jRubyRuntime);
        } catch (Exception e) {
            logger.log(Level.WARNING, "Error stopping application : "+contextRoot +". "+e.getMessage(), e);
        }
        return true;
    }

    /**
     * Suspends this application container.
     *
     * @return true if suspending was successful, false otherwise.
     */
    public boolean suspend() {
        // Not (yet) supported
        return false;
    }

    /**
     * Resumes this application container.
     *
     * @return true if resumption was successful, false otherwise.
     */
    public boolean resume() {
        // Not (yet) supported
        return false;
    }

    public ClassLoader getClassLoader() {
        return null;
    }

    /**
     * Returns the deployment descriptor associated with this application
     *
     * @return deployment descriptor if they exist or null if not
     */
    public Object getDescriptor() {
        return null;
    }

    public String getContextRoot() {
        return contextRoot;
    }

    public void service(Request request, Response response) throws Exception {
        adapter.service(request, response);
    }

    public void afterService(Request request, Response response) throws Exception {
        adapter.afterService(request, response);
    }

    public void fireAdapterEvent(String s, Object o) {
        adapter.fireAdapterEvent(s, o);
    }


    private static final class JRubyRailsClassLoader extends URLClassLoader {

        private final ClassLoader parent;
        public JRubyRailsClassLoader(URL[] urls, ClassLoader parent) {
            super(urls);
            this.parent = parent;
        }

        @Override
        public Class<?> loadClass(String name) throws ClassNotFoundException {

            Class<?> c;
            try {
                c = parent.loadClass(name);
            } catch (ClassNotFoundException e) {
                c = super.loadClass(name);
            }
            return c;
        }
    }

    private URL[] findJRubyJars() throws MalformedURLException, ClassNotFoundException {
        List<URL> urls = new ArrayList<URL>();

        File jrubyLibDir = new File(jrubyEnv.jrubyLib);
        File jrubyJar = new File(jrubyEnv.jrubyLib, "jruby.jar");

        if (!jrubyJar.exists()) {
            throw new IllegalArgumentException("jruby.jar at: " + jrubyEnv.jrubyLib + " not found.");
        }

        for (File lib : jrubyLibDir.listFiles()) {
            if (lib.isFile() && lib.toString().endsWith(".jar")) {
                urls.add(lib.toURI().toURL());
            }
        }

        //Add $GLASSFISH_HOME/modules
        File grizzlyjruby = env.getStartupContext().getRootDirectory();
        File[] grizzlyjrubyJars = grizzlyjruby.listFiles(new FilenameFilter(){
            public boolean accept(File dir, String name) {
                return !name.startsWith("grizzly-jruby-module") && name.startsWith("grizzly-jruby");
            }
        });

        //TODO: if there are more than one, print warning and add the latest version
        if(grizzlyjrubyJars.length > 0)
            urls.add(grizzlyjrubyJars[0].toURI().toURL());

        return urls.toArray(new URL[urls.size()]);
    }

    private final class JRubyEnvironment {
        final String jrubyHome = System.getProperty("jruby.home");
        final int numberOfRuntime;
        final String jrubyLib;

        private final String jrubyRuntime = System.getProperty("jruby.runtime");
        private final String railsEnv = System.getProperty("RAILS_ENV");

        public JRubyEnvironment(StartupContext startupContext) {
            if (jrubyHome == null) {
                logger.severe("jruby.home cannot be null");
                throw new IllegalArgumentException("jruby.home cannot be null");
            }

            // For now using the numberOfRuntimes provided by the user as is, later
            // this would be tied to the mode of deployment production/development/test
            // In the near future provide a jruby-container element in the domain.xml
            // that could have all the jruby related information as part of it.
            int numRt = 1;

            if (jrubyRuntime != null) {
                try {
                    numRt = Integer.parseInt(jrubyRuntime);
                } catch (NumberFormatException ex) {
                    logger.log(Level.WARNING, "Invalid number of Runtimes specified.");
                }
            } else if (railsEnv != null && railsEnv.equalsIgnoreCase("production")) {
                // By default if the user has defined a production environment and
                // not provided the number of runtimes to start we start with a
                // default value of '3'. Why '3' (have no idea, just a number that
                // I decided on).
                numRt = 3;
            }


            //check the commandline, above whatever we do is for legacy
            if(startupContext.getArguments().get("--runtimes") != null){
                try {
                    numRt = Integer.parseInt(startupContext.getArguments().get("--runtimes"));
                } catch (NumberFormatException ex) {
                    logger.log(Level.WARNING, "Invalid number of Runtimes specified.");
                }
            }

            numberOfRuntime = numRt;
            jrubyLib = jrubyHome + "/lib";
        }
    }
}
